;;; Directory Local Variables
;;; For more information see (info "(emacs) Directory Variables")

((sh-mode
  (sh-basic-offset . 8)
  (sh-indentation . 8)
  (sh-indent-for-case-label . 0)
  (sh-indent-for-case-alt . +)))
